************************************************************************
*
* PCG: Preconditioned Conjugate Gradient Package
* Version: f77
*
************************************************************************
************************************************************************
*# 
*#                      COPYRIGHT/DISCLAIMER NOTICE
*#
*# This program was prepared by the Regents of the University of
*# California at Los Alamos National Laboratory under Contract No.
*# W-7405-ENG-36 with the U.S. Department of Energy (DOE), and by
*# the University of Texas at Austin under ARPA Contract No.
*# DABT63-92-C-0024.
*# 
*# The University of California and the University of Texas at
*# Austin have certain rights in the program pursuant to these
*# contracts.
*# 
*# Permission is hereby granted to use the program for the user's
*# own internal use.  The user is not granted the right to reproduce,
*# prepare derivative works, or redistribute the program without
*# prior permission of the University of California or the University
*# of Texas at Austin.
*# 
*# All rights in the program are reserved by the University of
*# California and the University of Texas at Austin.
*# 
*# Portions of this material resulted from work developed under a
*# U.S. Government Contract and are subject to the following
*# license: the Government is granted for itself and others acting
*# on its behalf a paid-up, nonexclusive, irrevocable worldwide
*# license in this computer software to reproduce, prepare derivative
*# works, and perform publicly and display publicly.
*# 
*# Neither the U.S. Government, the University of California nor
*# the University of Texas at Austin, nor any of their employees,
*# makes any warranty, express or implied, or assumes any liability
*# or responsibility for the use of this software.
*# 
*# Copyright (c) 1992-1995 the University of California and the
*# University of Texas at Austin.
*# 
************************************************************************
************************************************************************
************************************************************************
**  ^ROUTINE: WCGNPR   - Reverse communication level routine for
**                     conjugate gradient acceleration, Neumann
**                     polynomial (left) preconditioning. (^)
**
************************************************************************
**
      subroutine scgnpr ( ijob , ireq ,
     &       u , uexact , b ,
     &       iva , ivql , ivqr ,
     &       iwk , fwk , iparm , fparm , ier )
**
************************************************************************
**
**  ^DESCRIPTION:
**    Standard conjugate gradient, (left) Neumann polynomial
**    preconditioned.  The standard preconditioning as in CG
**    acceleration is permitted, but then Neumann polynomial
**    preconditioning on the left is layerd onto that before
**    CG.
**
**  ^AUTHOR:   wdj@beta.lanl.gov
**
**  ^MODIFIED: wdj@lithos.c3.lanl.gov Thu May  6 13:12:38 MDT 1993
**
**  ^ARGUMENTS: see *Subroutine Arguments* below.
**
**  ^REQUIREMENTS:
**    Common Blocks: none
**    Subroutines:   see below.
**
**  ^SIDE_EFFECTS:
**    See the definition of METHRDCLS to see which parameters
**    are modified
**
**  ^ALGORITHM:
**   $$ p(z) = \alpha \sum_{i=0} ^k (I-\alpha z)^i $$
**   $$ d^{(n)}  = p(Q_R Q_L A) Q_R Q_L r^{(n)} $$
**   $$ p^{(0)}  = d^{(0)} $$
**   $$ p^{(n)}  = d^{(n)} + \alpha_n p^{(n-1)}, \qquad n>0 $$
**   $$ \alpha_n = {( r^{(n  )}, d^{(n  )}) \over
**                  ( r^{(n-1)}, d^{(n-1)})} $$
**   $$ u^{(n+1)} = u^{(n)} + \lambda_n   p^{(n)}
**      r^{(n+1)} = r^{(n)} - \lambda_n A p^{(n)} $$
**   $$ \lambda_n = {( r^{(n)},  d^{(n)}) \over
**                   ( p^{(n)}, Ap^{(n)})} $$
**
**
**  ^REFERENCES:
**   Steven F. Ashby, Thomas A. Manteuffel and Paul E. Saylor,
**   ``A Taxonomy for Conjugate Gradient Methods,'' {\sl SIAM J. Numer.
**   Anal.}, 27 (1990), pp. 1542--1568.
**
**#  O. G. Johnson, C. A. Micchelli, and G. Paul, ``Polynomial
**   Preconditioning for Conjugate Gradient Calculations,'' {\sl SIAM
**   J. Numer. Anal.}, 20 (1983), pp. 362--376.
**
**  ^DOCUMENTATION:
**
**  ^MACROS:
**
************************************************************************
*
*     implicit character*1 (a-z)
*
*#    METHRAL        - (^)
      integer ijob
      integer ireq
      integer ier
      integer iwk(*)
      integer iparm(*)
      real       fwk(*)
      real       fparm(*)
      real u(*)
      real uexact(*)
      real b(*)
      integer ivql
      integer iva
      integer ivqr
*
        integer itimer
        double precision tc1, tr1, tc2, tr2
      integer    iitab
      integer    iftab
      integer    iwkstg
      integer    iclstg
      integer    in
      integer    is
      integer    iretlb
      integer    ivitmp, ivotmp
      integer    iv1tst, iv2tst
      integer    istab1, istab2, istab
      integer    ib1,ib2, iu1,iu2
      integer    iud
      parameter (iud  =0)
      integer    ibd
      parameter (ibd  =1)
      integer    ilbd
      parameter (ilbd =2)
      integer    irlbd
      parameter (irlbd=3)
      integer    ied
      parameter (ied  =4)
      integer    ird
      parameter (ird  =5)
      integer    ilrd
      parameter (ilrd =6)
      integer    irlrd
      parameter (irlrd=7)
      real      urndf
      real      top
      real      bot
      parameter (urndf=.1)
      real      mydot1, mydot2
      external   sargck
      external   sckstg
      external   swrnd
      external   sifppr
      external   sitput
      external   xtimer
        integer ipme
        integer nproc
      external ximini
      external ximal
      external xifre
      external sfmini
      external sfmal
      external sffre
*     integer imv1, imv2
      integer    ir
      integer    ip
      integer    iap
      integer    iqlr
      integer    iqrqlr
      integer    iv1
      integer    iv2
      integer    iv3
      integer    idel
      integer    ns
      integer    i
      real      alphan
      real      pap
      real      rdot
      real      rnorm
      real      r0norm
      real      rr
      real      rrold
      real      vlamda
      real      alpha
      real      cval
      real      uudot, uexdt
*
*
****^^******************************************************************
*     $Modified: wdj@lanl.gov Mon Aug 28 19:20:55 MDT 1995
*     $Id: cgnpr.fm4,v 1.35 1994/11/22 05:20:14 joubert Exp $
*     $Revision: 1.0 $
************************************************************************
*
      iitab = (iparm(6))
      iftab = (iparm(6))
      ip = (iparm(6))
      iap = (iparm(6))
      ir = (iparm(6))
      iv1 = (iparm(6))
      iv2 = (iparm(6))
      iv3 = (iparm(6))
      idel = (iparm(6))
*
*     *---go to the brancher unless this is the init call---
      if (ijob .ne. 1) go to 1000
*
*     *---inits----------
      iclstg = 1
      itimer = 0
      in = 0
      iparm(5) = in
      is = 0
*
*     *---initialize pointers----------
      iu1 = (iparm(6))
      iu2 = (iparm(6))
      ib1 = (iparm(6))
      ib2 = (iparm(6))
      istab1 = (iparm(6))
      istab2 = (iparm(6))
*
*     *---initialize----------
      ipme  = 0
      nproc = 1
      if (iparm(27) .eq. 0) iparm(27) = 0
      if (iparm(27) .eq. 0) then
        ier = 0
        if (ijob.eq.1 .or. ijob.eq.2 .or. ijob.eq.0) then
*       *---init integer memory----------
        call ximini ( iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
*       *---init floating point memory----------
        call sfmini ( iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
        endif
*       *---check arguments----------
        call sargck ( iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
*       *---initialize timer----------
        call xtimer (tc1,tr1,iparm(22),1)
      endif
*
*     *---start timer------
      call xtimer (tc1,tr1,iparm(22),0)
      itimer = 1
*
*     *---print i/fparm if requested----------
      if (iparm(27) .eq. 0) then
        call sifppr ( 1 , iparm, fparm, ier)
           if (ier .lt. 0) go to 900
      endif
*
*     *---reserve place to save integer and float variables----------
*     *---across revcom calls----------
      call ximal (iitab,30, iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
      call sfmal (iftab,12, iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
*
      if (iparm(2) .ge. 3) then
        write (iparm(1),'(/''     Left Neumann Poly Prec Conjugate Gradi
     &ent''/)') 
      endif
*
      ns = iparm(16)
*
      alpha = fparm(6)
*
      call sfmal (ip ,iparm(3), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
      call sfmal (iap,iparm(3), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
      call sfmal (ir ,iparm(3), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
      if (.not.((iparm(13).eq.2).or.(iparm(13).eq.3))) then
        if (.not.((iparm(13).eq.1).or.(iparm(13).eq.3))) then
          iqlr   = ir
          iqrqlr = iqlr
        else
          iqlr   = iap
          iqrqlr = iqlr
        endif
      else
        if (.not.((iparm(13).eq.1).or.(iparm(13).eq.3))) then
          iqlr   = ir
          iqrqlr = iap
        else
          iqlr   = iap
          call sfmal (iqrqlr,iparm(3), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
        endif
      endif
      call sfmal (iv1,iparm(3), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
      call sfmal (iv2,iparm(3), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
      call sfmal (iv3,iparm(3), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
      call sfmal (idel,iparm(3), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
*
*
*     *---set u to zero----------
      if (iparm(14) .eq. -2
     &    .or. iparm(14) .eq. -1) then
          call swfil (u,0e0,
     &     iparm , fparm , iwk , fwk , ier  )
      endif
*
*     *---random u----------
      if (iparm(14) .eq. 2
     &    ) go to 5500
                 go to 5501
 5500  continue
          call swrdot (mydot1,b,
     &    b,  iparm , fparm , iwk , fwk , ier  )
        call swrnd (u,  iparm , fparm , iwk , fwk , ier  )
*        *---begin revcom matvec----------
          call sfmal (ivitmp,iparm(3), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
          ivqr = ivitmp
          call swset (fwk(ivqr),u,
     &     iparm , fparm , iwk , fwk , ier  )
          iva = ir 
         ireq = 3
        iretlb = 2000
        go to 1500
 2000   continue
          call sffre (ivitmp,iparm(3)*(1), iparm , fparm , iwk , fwk , i
     &er )
           if (ier .lt. 0) go to 900
         if (ijob .eq. -1) go to 900
*        *---end revcom matvec----------
          call swrdot (mydot2,fwk(ir),
     &    fwk(ir),  iparm , fparm , iwk , fwk , ier  )
          call swscl ( u, u, urndf*sqrt(mydot1/mydot2),  iparm , fparm ,
     & iwk , fwk , ier  )
                go to 5502
 5501 continue
 5502 continue
*
*     *---random u perturbation----------
      if (iparm(14) .eq. 3) go to 5503
                 go to 5504
 5503  continue
*        *---begin revcom matvec----------
          call sfmal (ivitmp,iparm(3), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
          ivqr = ivitmp
          call swset (fwk(ivqr),u,
     &     iparm , fparm , iwk , fwk , ier  )
          iva = ir 
         ireq = 3
        iretlb = 2001
        go to 1500
 2001   continue
          call sffre (ivitmp,iparm(3)*(1), iparm , fparm , iwk , fwk , i
     &er )
           if (ier .lt. 0) go to 900
         if (ijob .eq. -1) go to 900
*        *---end revcom matvec----------
          call swsub ( fwk(ir),
     &    b, fwk(ir),  iparm , fparm , iwk , fwk , ier  )
          call swrdot (mydot1,fwk(ir),
     &    fwk(ir),  iparm , fparm , iwk , fwk , ier  )
          call swset (fwk(ir),u,
     &     iparm , fparm , iwk , fwk , ier  )
        call swrnd (u,  iparm , fparm , iwk , fwk , ier  )
*        *---begin revcom matvec----------
          call sfmal (ivitmp,iparm(3), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
          ivqr = ivitmp
          call swset (fwk(ivqr),u,
     &     iparm , fparm , iwk , fwk , ier  )
          iva = ip 
         ireq = 3
        iretlb = 2002
        go to 1500
 2002   continue
          call sffre (ivitmp,iparm(3)*(1), iparm , fparm , iwk , fwk , i
     &er )
           if (ier .lt. 0) go to 900
         if (ijob .eq. -1) go to 900
*        *---end revcom matvec----------
          call swrdot (mydot2,fwk(ip),
     &    fwk(ip),  iparm , fparm , iwk , fwk , ier  )
          call swsax (u,fwk(ir),
     &    u, -urndf*sqrt(mydot1/mydot2),  iparm , fparm , iwk , fwk , ie
     &r  )
                go to 5505
 5504 continue
 5505 continue
*
*     *---begin iteration statistics calculation----------
      if (iparm(21) .eq. 1) go to 5506
                 go to 5507
 5506  continue
        call sfmal (istab1,8, iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
        call sfmal (istab2,8, iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
        call sfmal (ib1,iparm(3), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
        call sfmal (ib2,iparm(3), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
        call sfmal (iu1,iparm(3), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
        call sfmal (iu2,iparm(3), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
        istab = istab1
        if (iparm(19) .eq. 1) then
          call swdot (fwk(istab+iud), uexact,
     &    uexact,  iparm , fparm , iwk , fwk , ier  )
        endif
          call swdot (fwk(istab+ibd ), b,
     &    b,  iparm , fparm , iwk , fwk , ier  )
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.1).or.(iparm(13).eq.3))) go to 5509
                 go to 5510
 5509  continue
          call swset (fwk(ib2),b,
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5511
 5510 continue
          call sfmal (ivitmp,iparm(3), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
          iva = ivitmp
          call swset (fwk(iva),b,
     &     iparm , fparm , iwk , fwk , ier  )
          ivql = ib2 
         ireq = 5
        iretlb = 2003
        go to 1500
 2003   continue
          call sffre (ivitmp,iparm(3)*(1), iparm , fparm , iwk , fwk , i
     &er )
           if (ier .lt. 0) go to 900
         if (ijob .eq. -1) go to 900
 5511 continue
*        *---end revcom precon----------
          call swdot (fwk(istab+ilbd), fwk(ib2),
     &    fwk(ib2),  iparm , fparm , iwk , fwk , ier  )
          call swset (fwk(iu1),fwk(ib2),
     &     iparm , fparm , iwk , fwk , ier  )
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.2).or.(iparm(13).eq.3))) go to 5512
                 go to 5513
 5512  continue
          call swset (fwk(iu2),fwk(iu1),
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5514
 5513 continue
          ivql = iu1 
          ivqr = iu2 
         ireq = 7
        iretlb = 2004
        go to 1500
 2004   continue
         if (ijob .eq. -1) go to 900
 5514 continue
*        *---end revcom precon----------
          call swdot (fwk(istab+irlbd), fwk(iu2),
     &    fwk(iu2),  iparm , fparm , iwk , fwk , ier  )
        if (iparm(19) .eq. 1) then
          call swsub ( fwk(iu1),
     &    u, uexact,  iparm , fparm , iwk , fwk , ier  )
          call swdot (fwk(istab+ied), fwk(iu1),
     &    fwk(iu1),  iparm , fparm , iwk , fwk , ier  )
        endif
*        *---begin revcom matvec----------
          call sfmal (ivitmp,iparm(3), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
          ivqr = ivitmp
          call swset (fwk(ivqr),u,
     &     iparm , fparm , iwk , fwk , ier  )
          iva = ib1 
         ireq = 3
        iretlb = 2005
        go to 1500
 2005   continue
          call sffre (ivitmp,iparm(3)*(1), iparm , fparm , iwk , fwk , i
     &er )
           if (ier .lt. 0) go to 900
         if (ijob .eq. -1) go to 900
*        *---end revcom matvec----------
          call swsub ( fwk(ib1),
     &    b, fwk(ib1),  iparm , fparm , iwk , fwk , ier  )
          call swdot (fwk(istab+ird), fwk(ib1),
     &    fwk(ib1),  iparm , fparm , iwk , fwk , ier  )
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.1).or.(iparm(13).eq.3))) go to 5515
                 go to 5516
 5515  continue
          call swset (fwk(ib2),fwk(ib1),
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5517
 5516 continue
          iva = ib1 
          ivql = ib2 
         ireq = 5
        iretlb = 2006
        go to 1500
 2006   continue
         if (ijob .eq. -1) go to 900
 5517 continue
*        *---end revcom precon----------
          call swdot (fwk(istab+ilrd ), fwk(ib2),
     &    fwk(ib2),  iparm , fparm , iwk , fwk , ier  )
          call swset (fwk(iu1),fwk(ib2),
     &     iparm , fparm , iwk , fwk , ier  )
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.2).or.(iparm(13).eq.3))) go to 5518
                 go to 5519
 5518  continue
          call swset (fwk(iu2),fwk(iu1),
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5520
 5519 continue
          ivql = iu1 
          ivqr = iu2 
         ireq = 7
        iretlb = 2007
        go to 1500
 2007   continue
         if (ijob .eq. -1) go to 900
 5520 continue
*        *---end revcom precon----------
          call swdot (fwk(istab+irlrd), fwk(iu2),
     &    fwk(iu2),  iparm , fparm , iwk , fwk , ier  )
        if (iparm(2) .ge. 4) then
        write (iparm(1),'()') 
        write (iparm(1),'(1x, 60(''*''))') 
        write (iparm(1),'()') 
        write (iparm(1),'(5x,''Initial iteration statistics'')') 
        write (iparm(1),'()') 
        if (iparm(19) .eq. 1) then
        write (iparm(1),'(5x,''2-norm of uexact            '',1pg20.10)'
     &) sqrt(fwk(istab+iud))
        endif
        write (iparm(1),'(5x,''2-norm of b                 '',1pg20.10)'
     &) sqrt(fwk(istab+ibd))
        write (iparm(1),'(5x,''2-norm of Ql*b              '',1pg20.10)'
     &) sqrt(fwk(istab+ilbd))
        write (iparm(1),'(5x,''2-norm of Qr*Ql*b           '',1pg20.10)'
     &) sqrt(fwk(istab+irlbd))
        if (iparm(19) .eq. 1) then
        write (iparm(1),'(5x,''2-norm of error             '',1pg20.10)'
     &) sqrt(fwk(istab+ied))
        endif
        write (iparm(1),'(5x,''2-norm of residual          '',1pg20.10)'
     &) sqrt(fwk(istab+ird))
        write (iparm(1),'(5x,''2-norm of Ql*r              '',1pg20.10)'
     &) sqrt(fwk(istab+ilrd))
        write (iparm(1),'(5x,''2-norm of Qr*Ql*r           '',1pg20.10)'
     &) sqrt(fwk(istab+irlrd))
        write (iparm(1),'()') 
        write (iparm(1),'(1x, 60(''*''))') 
        write (iparm(1),'()') 
        endif
        call sffre (iu2,iparm(3)*(1), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
        call sffre (iu1,iparm(3)*(1), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
        call sffre (ib2,iparm(3)*(1), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
        call sffre (ib1,iparm(3)*(1), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
                go to 5508
 5507 continue
 5508 continue
*     *---end iteration statistics calculation----------
*
      if (((iparm(14).eq.0).or.(iparm(14).eq.-2)) .or. iparm(14).eq.-1) 
     &go to 5521
                 go to 5522
 5521  continue
          call swset (fwk(ir),b,
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5523
 5522 continue
*        *---begin revcom matvec----------
          call sfmal (ivitmp,iparm(3), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
          ivqr = ivitmp
          call swset (fwk(ivqr),u,
     &     iparm , fparm , iwk , fwk , ier  )
          iva = ir 
         ireq = 3
        iretlb = 2008
        go to 1500
 2008   continue
          call sffre (ivitmp,iparm(3)*(1), iparm , fparm , iwk , fwk , i
     &er )
           if (ier .lt. 0) go to 900
         if (ijob .eq. -1) go to 900
*        *---end revcom matvec----------
          call swsub ( fwk(ir),
     &    b, fwk(ir),  iparm , fparm , iwk , fwk , ier  )
 5523 continue
*
      if (((iparm(13).eq.1).or.(iparm(13).eq.3))) go to 5524
                 go to 5525
 5524  continue
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.1).or.(iparm(13).eq.3))) go to 5527
                 go to 5528
 5527  continue
          call swset (fwk(iqlr),fwk(ir),
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5529
 5528 continue
          iva = ir 
          ivql = iqlr 
         ireq = 5
        iretlb = 2009
        go to 1500
 2009   continue
         if (ijob .eq. -1) go to 900
 5529 continue
*        *---end revcom precon----------
                go to 5526
 5525 continue
 5526 continue
      if (((iparm(13).eq.2).or.(iparm(13).eq.3))) go to 5530
                 go to 5531
 5530  continue
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.2).or.(iparm(13).eq.3))) go to 5533
                 go to 5534
 5533  continue
          call swset (fwk(iqrqlr),fwk(iqlr),
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5535
 5534 continue
          ivql = iqlr 
          ivqr = iqrqlr 
         ireq = 7
        iretlb = 2010
        go to 1500
 2010   continue
         if (ijob .eq. -1) go to 900
 5535 continue
*        *---end revcom precon----------
                go to 5532
 5531 continue
 5532 continue
*
          call swrdot (rdot,fwk(iqlr),
     &    fwk(iqlr),  iparm , fparm , iwk , fwk , ier  )
      rnorm  = sqrt(abs(rdot))
      r0norm = rnorm
*
*---------------------------iteration loop------------------------------
*
 100  continue
*
        if (iparm(2) .ge. 5) then
        write (iparm(1),500) '(p(QR QL A) QR QL)^1/2 r norm  ', rnorm
        endif
        if (iparm(2) .ge. 5) then
        write (iparm(1),500) '((p(QR QL A) QR QL) *r,r)      ', rr
        endif
*
*     *---begin user stopping test----------
        if (iparm(12).eq.-3) go to 5536
                 go to 5537
 5536  continue
          if ((mod(iparm(15)/1,2).eq.1)) then
          endif
          if ((mod(iparm(15)/2,2).eq.1)) then
          iva = ir 
          endif
          if ((mod(iparm(15)/4,2).eq.1)) then
          ivql = iqlr 
          endif
          if ((mod(iparm(15)/8,2).eq.1)) then
          ivqr = iqrqlr 
          endif
          ireq   = 9
        iretlb = 2011
        go to 1500
 2011   continue
          if ((mod(iparm(15)/8,2).eq.1)) then
          endif
          if ((mod(iparm(15)/4,2).eq.1)) then
          endif
          if ((mod(iparm(15)/2,2).eq.1)) then
          endif
          if ((mod(iparm(15)/1,2).eq.1)) then
          endif
          if (ier .ge. 0) then
            call sitput ( in , is ,  iparm , fparm , iwk , fwk , ier )
          endif
          if (ijob .eq. -1) then
            if (iparm(2).ge.3 .and. iparm(27).eq.0) then
        write (iparm(1),'(/a/)') ' User-requested termination of iterati
     &ve method.'
            endif
            go to 900
          endif
                go to 5538
 5537 continue
 5538 continue
        if (iparm(12) .gt. 1) go to 5539
                 go to 5540
 5539  continue
          if (iparm(12) .eq. 1 ) then
            if (iparm(5) .eq. 0) then
              if (iparm(19) .eq. 0) then
                ier = -5
                call xersho ( ier, 'scgnpr' , iparm ,
     &                        '19 and 12 incompatible' )
                go to 900
              endif
          call swdot (bot, uexact,
     &    uexact,  iparm , fparm , iwk , fwk , ier  )
            endif
            call sfmal (iv1tst,iparm(3), iparm , fparm , iwk , fwk , ier
     & )
           if (ier .lt. 0) go to 900
          call swsub ( fwk(iv1tst),
     &    u, uexact,  iparm , fparm , iwk , fwk , ier  )
          call swdot (top, fwk(iv1tst),
     &    fwk(iv1tst),  iparm , fparm , iwk , fwk , ier  )
            call sffre (iv1tst,iparm(3)*(1), iparm , fparm , iwk , fwk ,
     & ier )
           if (ier .lt. 0) go to 900
          endif
          if (iparm(12) .eq. 2 ) then
            if (iparm(5) .eq. 0) then
          call swdot (bot, b,
     &    b,  iparm , fparm , iwk , fwk , ier  )
            endif
          call swdot (top, fwk(ir),
     &    fwk(ir),  iparm , fparm , iwk , fwk , ier  )
          endif
          if (iparm(12) .eq. 3) go to 5542
                 go to 5543
 5542  continue
            if (iparm(5) .eq. 0) go to 5545
                 go to 5546
 5545  continue
              if (((iparm(14).eq.0).or.(iparm(14).eq.-2))) go to 5548
                 go to 5549
 5548  continue
          call swdot (bot, fwk(iqlr),
     &    fwk(iqlr),  iparm , fparm , iwk , fwk , ier  )
                go to 5550
 5549 continue
                call sfmal (iv1tst,iparm(3), iparm , fparm , iwk , fwk ,
     & ier )
           if (ier .lt. 0) go to 900
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.1).or.(iparm(13).eq.3))) go to 5551
                 go to 5552
 5551  continue
          call swset (fwk(iv1tst),b,
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5553
 5552 continue
          call sfmal (ivitmp,iparm(3), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
          iva = ivitmp
          call swset (fwk(iva),b,
     &     iparm , fparm , iwk , fwk , ier  )
          ivql = iv1tst 
         ireq = 5
        iretlb = 2012
        go to 1500
 2012   continue
          call sffre (ivitmp,iparm(3)*(1), iparm , fparm , iwk , fwk , i
     &er )
           if (ier .lt. 0) go to 900
         if (ijob .eq. -1) go to 900
 5553 continue
*        *---end revcom precon----------
          call swdot (bot, fwk(iv1tst),
     &    fwk(iv1tst),  iparm , fparm , iwk , fwk , ier  )
                call sffre (iv1tst,iparm(3)*(1), iparm , fparm , iwk , f
     &wk , ier )
           if (ier .lt. 0) go to 900
 5550 continue
                go to 5547
 5546 continue
 5547 continue
          call swdot (top, fwk(iqlr),
     &    fwk(iqlr),  iparm , fparm , iwk , fwk , ier  )
                go to 5544
 5543 continue
 5544 continue
          if (iparm(12) .eq. 4) go to 5554
                 go to 5555
 5554  continue
            if (iparm(5) .eq. 0) go to 5557
                 go to 5558
 5557  continue
              if (((iparm(14).eq.0).or.(iparm(14).eq.-2))) go to 5560
                 go to 5561
 5560  continue
          call swdot (bot, fwk(iqrqlr),
     &    fwk(iqrqlr),  iparm , fparm , iwk , fwk , ier  )
                go to 5562
 5561 continue
                call sfmal (iv1tst,iparm(3), iparm , fparm , iwk , fwk ,
     & ier )
           if (ier .lt. 0) go to 900
                call sfmal (iv2tst,iparm(3), iparm , fparm , iwk , fwk ,
     & ier )
           if (ier .lt. 0) go to 900
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.1).or.(iparm(13).eq.3))) go to 5563
                 go to 5564
 5563  continue
          call swset (fwk(iv1tst),b,
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5565
 5564 continue
          call sfmal (ivitmp,iparm(3), iparm , fparm , iwk , fwk , ier )
           if (ier .lt. 0) go to 900
          iva = ivitmp
          call swset (fwk(iva),b,
     &     iparm , fparm , iwk , fwk , ier  )
          ivql = iv1tst 
         ireq = 5
        iretlb = 2013
        go to 1500
 2013   continue
          call sffre (ivitmp,iparm(3)*(1), iparm , fparm , iwk , fwk , i
     &er )
           if (ier .lt. 0) go to 900
         if (ijob .eq. -1) go to 900
 5565 continue
*        *---end revcom precon----------
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.2).or.(iparm(13).eq.3))) go to 5566
                 go to 5567
 5566  continue
          call swset (fwk(iv2tst),fwk(iv1tst),
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5568
 5567 continue
          ivql = iv1tst 
          ivqr = iv2tst 
         ireq = 7
        iretlb = 2014
        go to 1500
 2014   continue
         if (ijob .eq. -1) go to 900
 5568 continue
*        *---end revcom precon----------
          call swdot (bot, fwk(iv2tst),
     &    fwk(iv2tst),  iparm , fparm , iwk , fwk , ier  )
                call sffre (iv2tst,iparm(3)*(1), iparm , fparm , iwk , f
     &wk , ier )
           if (ier .lt. 0) go to 900
                call sffre (iv1tst,iparm(3)*(1), iparm , fparm , iwk , f
     &wk , ier )
           if (ier .lt. 0) go to 900
 5562 continue
                go to 5559
 5558 continue
 5559 continue
          call swdot (top, fwk(iqrqlr),
     &    fwk(iqrqlr),  iparm , fparm , iwk , fwk , ier  )
                go to 5556
 5555 continue
 5556 continue
          if (iparm(12) .eq. 1 ) then
            if (iparm(19) .eq. 0) then
                ier = -5
                call xersho ( ier, 'scgnpr' , iparm ,
     &                        '19 and 12 incompatible' )
                go to 900
            endif
            if (((iparm(14).eq.0).or.(iparm(14).eq.-2))) then
          call swdot (top, uexact,
     &    uexact,  iparm , fparm , iwk , fwk , ier  )
            else
              call sfmal (iv1tst,iparm(3), iparm , fparm , iwk , fwk , i
     &er )
           if (ier .lt. 0) go to 900
          call swsub ( fwk(iv1tst),
     &    u, uexact,  iparm , fparm , iwk , fwk , ier  )
          call swdot (top, fwk(iv1tst),
     &    fwk(iv1tst),  iparm , fparm , iwk , fwk , ier  )
              call sffre (iv1tst,iparm(3)*(1), iparm , fparm , iwk , fwk
     & , ier )
           if (ier .lt. 0) go to 900
            endif
            if (iparm(5) .eq. 0) bot = top
          endif
          if (iparm(12) .eq. 6) then
          call swdot (top, fwk(ir),
     &    fwk(ir),  iparm , fparm , iwk , fwk , ier  )
            if (iparm(5) .eq. 0) bot = top
          endif
          if (iparm(12) .eq. 7) then
          call swdot (top, fwk(iqlr),
     &    fwk(iqlr),  iparm , fparm , iwk , fwk , ier  )
            if (iparm(5) .eq. 0) bot = top
          endif
          if (iparm(12) .eq. 8) then
          call swdot (top, fwk(iqrqlr),
     &    fwk(iqrqlr),  iparm , fparm , iwk , fwk , ier  )
            if (iparm(5) .eq. 0) bot = top
          endif
          if ((top) .lt. (0e0)) then
                ier = -10
                call xersho ( ier, 'scgnpr' , iparm ,
     &                        'Attempt to take sqrt of negative number' 
     &)
                go to 900
          endif
          if ((bot) .le. (0e0)) then
                ier = -10
                call xersho ( ier, 'scgnpr' , iparm ,
     &                        'Attempt to divide by zero' )
                go to 900
          endif
          fparm(5) = sqrt((top)/(bot))
          call sitput ( in , is ,  iparm , fparm , iwk , fwk , ier )
          if ((fparm(5)) .le. (fparm(4))) go to 2500
                go to 5541
 5540 continue
 5541 continue
*     *---end user stopping test----------
*
        if (iparm(12) .eq. -1) then
          if (r0norm .eq. 0e0) then
                ier = -10
                call xersho ( ier, 'scgnpr' , iparm ,
     &                        'Attempt to divide by zero' )
                go to 900
          endif
          fparm(5) = rnorm/r0norm
          call sitput ( in , is ,  iparm , fparm , iwk , fwk , ier )
          if ((fparm(5)) .le. (fparm(4))) go to 2500
        endif
*
*       *---begin exact stopping test----------
        if (iparm(12) .eq. -2) then
          call swsub ( fwk(iv1),
     &    u, uexact,  iparm , fparm , iwk , fwk , ier  )
          call swrdot (uudot,fwk(iv1),
     &    fwk(iv1),  iparm , fparm , iwk , fwk , ier  )
          call swrdot (uexdt,uexact,
     &    uexact,  iparm , fparm , iwk , fwk , ier  )
          if ((uudot)  .lt. (0e0)) then
                ier = -10
                call xersho ( ier, 'scgnpr' , iparm ,
     &                        'Attempt to take sqrt of negative number' 
     &)
                go to 900
          endif
          if ((uexdt) .le. (0e0)) then
                ier = -10
                call xersho ( ier, 'scgnpr' , iparm ,
     &                        'Attempt to divide by zero' )
                go to 900
          endif
          fparm(5) = sqrt((uudot)/(uexdt))
          call sitput ( in , is ,  iparm , fparm , iwk , fwk , ier )
          if ((fparm(5)) .le. (fparm(4))) go to 2500
        endif
*       *---end exact stopping test----------
*
        if (in .ge. iparm(4)) then
                ier = 2
                call xersho ( ier, 'scgnpr' , iparm , ' ' )
          go to 900
        endif
*
        call sckstg ( iclstg, iwkstg,  iparm , fparm , iwk , fwk , ier  
     &)
           if (ier .lt. 0) go to 900
*
*-----------------------proceed with iteration--------------------------
*
        if (ns .eq. 0) then
          call swscl ( fwk(idel), fwk(iqrqlr), alpha,  iparm , fparm , i
     &wk , fwk , ier  )
        endif
        if (ns.gt.0 .and. ns.le.3) go to 5569
                 go to 5570
 5569  continue
          cval = -(-alpha)**(ns+1)
          call swscl ( fwk(idel), fwk(iqrqlr), cval,  iparm , fparm , iw
     &k , fwk , ier  )
          i = 1
5572  continue
      if (i .gt. ns) go to 5573
            if (.not.((iparm(13).eq.2).or.(iparm(13).eq.3))) go to 5574
                 go to 5575
 5574  continue
              if (.not.((iparm(13).eq.1).or.(iparm(13).eq.3))) go to 557
     &7
                 go to 5578
 5577  continue
*        *---begin revcom matvec----------
          ivqr = idel 
          iva = iv1 
         ireq = 3
        iretlb = 2015
        go to 1500
 2015   continue
         if (ijob .eq. -1) go to 900
*        *---end revcom matvec----------
                go to 5579
 5578 continue
*        *---begin revcom matvec----------
          ivqr = idel 
          iva = iv2 
         ireq = 3
        iretlb = 2016
        go to 1500
 2016   continue
         if (ijob .eq. -1) go to 900
*        *---end revcom matvec----------
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.1).or.(iparm(13).eq.3))) go to 5580
                 go to 5581
 5580  continue
          call swset (fwk(iv1),fwk(iv2),
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5582
 5581 continue
          iva = iv2 
          ivql = iv1 
         ireq = 5
        iretlb = 2017
        go to 1500
 2017   continue
         if (ijob .eq. -1) go to 900
 5582 continue
*        *---end revcom precon----------
 5579 continue
                go to 5576
 5575 continue
              if (.not.((iparm(13).eq.1).or.(iparm(13).eq.3))) go to 558
     &3
                 go to 5584
 5583  continue
*        *---begin revcom matvec----------
          ivqr = idel 
          iva = iv2 
         ireq = 3
        iretlb = 2018
        go to 1500
 2018   continue
         if (ijob .eq. -1) go to 900
*        *---end revcom matvec----------
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.2).or.(iparm(13).eq.3))) go to 5586
                 go to 5587
 5586  continue
          call swset (fwk(iv1),fwk(iv2),
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5588
 5587 continue
          ivql = iv2 
          ivqr = iv1 
         ireq = 7
        iretlb = 2019
        go to 1500
 2019   continue
         if (ijob .eq. -1) go to 900
 5588 continue
*        *---end revcom precon----------
                go to 5585
 5584 continue
*        *---begin revcom matvec----------
          ivqr = idel 
          iva = iv1 
         ireq = 3
        iretlb = 2020
        go to 1500
 2020   continue
         if (ijob .eq. -1) go to 900
*        *---end revcom matvec----------
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.1).or.(iparm(13).eq.3))) go to 5589
                 go to 5590
 5589  continue
          call swset (fwk(iv2),fwk(iv1),
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5591
 5590 continue
          iva = iv1 
          ivql = iv2 
         ireq = 5
        iretlb = 2021
        go to 1500
 2021   continue
         if (ijob .eq. -1) go to 900
 5591 continue
*        *---end revcom precon----------
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.2).or.(iparm(13).eq.3))) go to 5592
                 go to 5593
 5592  continue
          call swset (fwk(iv1),fwk(iv2),
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5594
 5593 continue
          ivql = iv2 
          ivqr = iv1 
         ireq = 7
        iretlb = 2022
        go to 1500
 2022   continue
         if (ijob .eq. -1) go to 900
 5594 continue
*        *---end revcom precon----------
 5585 continue
 5576 continue
            if (alpha .ne. 0e0) cval = -cval*(ns+2-i)/alpha/i
          call swsax (fwk(idel),fwk(iv1),
     &    fwk(iqrqlr), cval,  iparm , fparm , iwk , fwk , ier  )
          i = i + 1
      go to 5572
5573  continue
                go to 5571
 5570 continue
 5571 continue
        if (ns.gt.3) go to 5595
                 go to 5596
 5595  continue
          call swscl ( fwk(idel), fwk(iqrqlr), alpha,  iparm , fparm , i
     &wk , fwk , ier  )
          if (ns .gt. 0) then
          call swset (fwk(iv3),fwk(idel),
     &     iparm , fparm , iwk , fwk , ier  )
          endif
          i = 1
5598  continue
      if (i .gt. ns) go to 5599
            if (.not.((iparm(13).eq.2).or.(iparm(13).eq.3))) go to 5600
                 go to 5601
 5600  continue
              if (.not.((iparm(13).eq.1).or.(iparm(13).eq.3))) go to 560
     &3
                 go to 5604
 5603  continue
*        *---begin revcom matvec----------
          ivqr = iv3 
          iva = iv1 
         ireq = 3
        iretlb = 2023
        go to 1500
 2023   continue
         if (ijob .eq. -1) go to 900
*        *---end revcom matvec----------
                go to 5605
 5604 continue
*        *---begin revcom matvec----------
          ivqr = iv3 
          iva = iv2 
         ireq = 3
        iretlb = 2024
        go to 1500
 2024   continue
         if (ijob .eq. -1) go to 900
*        *---end revcom matvec----------
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.1).or.(iparm(13).eq.3))) go to 5606
                 go to 5607
 5606  continue
          call swset (fwk(iv1),fwk(iv2),
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5608
 5607 continue
          iva = iv2 
          ivql = iv1 
         ireq = 5
        iretlb = 2025
        go to 1500
 2025   continue
         if (ijob .eq. -1) go to 900
 5608 continue
*        *---end revcom precon----------
 5605 continue
                go to 5602
 5601 continue
              if (.not.((iparm(13).eq.1).or.(iparm(13).eq.3))) go to 560
     &9
                 go to 5610
 5609  continue
*        *---begin revcom matvec----------
          ivqr = iv3 
          iva = iv2 
         ireq = 3
        iretlb = 2026
        go to 1500
 2026   continue
         if (ijob .eq. -1) go to 900
*        *---end revcom matvec----------
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.2).or.(iparm(13).eq.3))) go to 5612
                 go to 5613
 5612  continue
          call swset (fwk(iv1),fwk(iv2),
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5614
 5613 continue
          ivql = iv2 
          ivqr = iv1 
         ireq = 7
        iretlb = 2027
        go to 1500
 2027   continue
         if (ijob .eq. -1) go to 900
 5614 continue
*        *---end revcom precon----------
                go to 5611
 5610 continue
*        *---begin revcom matvec----------
          ivqr = iv3 
          iva = iv1 
         ireq = 3
        iretlb = 2028
        go to 1500
 2028   continue
         if (ijob .eq. -1) go to 900
*        *---end revcom matvec----------
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.1).or.(iparm(13).eq.3))) go to 5615
                 go to 5616
 5615  continue
          call swset (fwk(iv2),fwk(iv1),
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5617
 5616 continue
          iva = iv1 
          ivql = iv2 
         ireq = 5
        iretlb = 2029
        go to 1500
 2029   continue
         if (ijob .eq. -1) go to 900
 5617 continue
*        *---end revcom precon----------
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.2).or.(iparm(13).eq.3))) go to 5618
                 go to 5619
 5618  continue
          call swset (fwk(iv1),fwk(iv2),
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5620
 5619 continue
          ivql = iv2 
          ivqr = iv1 
         ireq = 7
        iretlb = 2030
        go to 1500
 2030   continue
         if (ijob .eq. -1) go to 900
 5620 continue
*        *---end revcom precon----------
 5611 continue
 5602 continue
          call swsax (fwk(iv3),fwk(iv3),
     &    fwk(iv1), -alpha,  iparm , fparm , iwk , fwk , ier  )
          call swadd (fwk(idel),
     &    fwk(idel), fwk(iv3),  iparm , fparm , iwk , fwk , ier  )
          i = i + 1
      go to 5598
5599  continue
                go to 5597
 5596 continue
 5597 continue
        if (in .ne. 0) rrold = rr
          call swrdot (rr,fwk(idel),
     &    fwk(ir),  iparm , fparm , iwk , fwk , ier  )
        if (in .eq. 0) then
          call swset (fwk(ip),fwk(idel),
     &     iparm , fparm , iwk , fwk , ier  )
        else
          if (rrold .eq. 0e0) then
                ier = -7
                call xersho ( ier, 'scgnpr' , iparm , ' ' )
                go to 900
          endif
          alphan = rr / rrold
          call swsax (fwk(ip),fwk(idel),
     &    fwk(ip), alphan,  iparm , fparm , iwk , fwk , ier  )
          if (iparm(2) .ge. 5) then
        write (iparm(1),500) '((p(QR QL A) QR QL) *r,r)     ', rr
          endif
        endif
*
*        *---begin revcom matvec----------
          ivqr = ip 
          iva = iap 
         ireq = 3
        iretlb = 2031
        go to 1500
 2031   continue
         if (ijob .eq. -1) go to 900
*        *---end revcom matvec----------
*
          call swcdot (pap, fwk(ip),
     &    fwk(iap), iparm(20),  iparm , fparm , iwk , fwk , ier  )
        if (iparm(2) .ge. 5) then
        write (iparm(1),500) '(p,ap)         ', pap
        endif
        if (pap .eq. 0e0) then
                ier = -6
                call xersho ( ier, 'scgnpr' , iparm , ' ' )
                go to 900
        endif
        vlamda = rr / pap
        if (iparm(2) .ge. 5) then
        write (iparm(1),500) 'lambda                     ', vlamda
        endif
*
          call swsax (u,u,
     &    fwk(ip), vlamda ,  iparm , fparm , iwk , fwk , ier  )
          call swsax (fwk(ir),fwk(ir),
     &    fwk(iap), (-vlamda),  iparm , fparm , iwk , fwk , ier  )
        if (((iparm(13).eq.1).or.(iparm(13).eq.3))) go to 5621
                 go to 5622
 5621  continue
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.1).or.(iparm(13).eq.3))) go to 5624
                 go to 5625
 5624  continue
          call swset (fwk(iqlr),fwk(ir),
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5626
 5625 continue
          iva = ir 
          ivql = iqlr 
         ireq = 5
        iretlb = 2032
        go to 1500
 2032   continue
         if (ijob .eq. -1) go to 900
 5626 continue
*        *---end revcom precon----------
                go to 5623
 5622 continue
 5623 continue
        if (((iparm(13).eq.2).or.(iparm(13).eq.3))) go to 5627
                 go to 5628
 5627  continue
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.2).or.(iparm(13).eq.3))) go to 5630
                 go to 5631
 5630  continue
          call swset (fwk(iqrqlr),fwk(iqlr),
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5632
 5631 continue
          ivql = iqlr 
          ivqr = iqrqlr 
         ireq = 7
        iretlb = 2033
        go to 1500
 2033   continue
         if (ijob .eq. -1) go to 900
 5632 continue
*        *---end revcom precon----------
                go to 5629
 5628 continue
 5629 continue
*
          call swrdot (rdot,fwk(iqlr),
     &    fwk(iqlr),  iparm , fparm , iwk , fwk , ier  )
        rnorm  = sqrt(abs(rdot))
*
        in         = in + 1
        iparm(5) = in
        is         = is + 1
*
      go to 100
*
 2500 continue
      if (iparm(2) .ge. 3) then
        write (iparm(1),'(/a,i7,a/)') '     Conjugate Gradient Method ha
     &s converged in ', iparm(5), ' iterations.'
      endif
*
*-------------------------------terminate-------------------------------
*
 900  continue
*
      if (iclstg .ne. 1) then
        call sckstg (-1, iwkstg,  iparm , fparm , iwk , fwk , ier )
      endif
*
*     *---begin iteration statistics calculation----------
      if (iparm(21).eq.1 .and. ier.ge.0) go to 5633
                 go to 5634
 5633  continue
        call sfmal (ib1,iparm(3), iparm , fparm , iwk , fwk , ier )
        call sfmal (ib2,iparm(3), iparm , fparm , iwk , fwk , ier )
        call sfmal (iu1,iparm(3), iparm , fparm , iwk , fwk , ier )
        call sfmal (iu2,iparm(3), iparm , fparm , iwk , fwk , ier )
        istab = istab2
        if (iparm(19) .eq. 1) then
          call swdot (fwk(istab+iud), uexact,
     &    uexact,  iparm , fparm , iwk , fwk , ier  )
        endif
          call swdot (fwk(istab+ibd ), b,
     &    b,  iparm , fparm , iwk , fwk , ier  )
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.1).or.(iparm(13).eq.3))) go to 5636
                 go to 5637
 5636  continue
          call swset (fwk(ib2),b,
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5638
 5637 continue
          call sfmal (ivitmp,iparm(3), iparm , fparm , iwk , fwk , ier )
          iva = ivitmp
          call swset (fwk(iva),b,
     &     iparm , fparm , iwk , fwk , ier  )
          ivql = ib2 
         ireq = 5
        iretlb = 2034
        go to 1500
 2034   continue
          call sffre (ivitmp,iparm(3)*(1), iparm , fparm , iwk , fwk , i
     &er )
         if (ijob .eq. -1) go to 900
 5638 continue
*        *---end revcom precon----------
          call swdot (fwk(istab+ilbd), fwk(ib2),
     &    fwk(ib2),  iparm , fparm , iwk , fwk , ier  )
          call swset (fwk(iu1),fwk(ib2),
     &     iparm , fparm , iwk , fwk , ier  )
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.2).or.(iparm(13).eq.3))) go to 5639
                 go to 5640
 5639  continue
          call swset (fwk(iu2),fwk(iu1),
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5641
 5640 continue
          ivql = iu1 
          ivqr = iu2 
         ireq = 7
        iretlb = 2035
        go to 1500
 2035   continue
         if (ijob .eq. -1) go to 900
 5641 continue
*        *---end revcom precon----------
          call swdot (fwk(istab+irlbd), fwk(iu2),
     &    fwk(iu2),  iparm , fparm , iwk , fwk , ier  )
        if (iparm(19) .eq. 1) then
          call swsub ( fwk(iu1),
     &    u, uexact,  iparm , fparm , iwk , fwk , ier  )
          call swdot (fwk(istab+ied), fwk(iu1),
     &    fwk(iu1),  iparm , fparm , iwk , fwk , ier  )
        endif
*        *---begin revcom matvec----------
          call sfmal (ivitmp,iparm(3), iparm , fparm , iwk , fwk , ier )
          ivqr = ivitmp
          call swset (fwk(ivqr),u,
     &     iparm , fparm , iwk , fwk , ier  )
          iva = ib1 
         ireq = 3
        iretlb = 2036
        go to 1500
 2036   continue
          call sffre (ivitmp,iparm(3)*(1), iparm , fparm , iwk , fwk , i
     &er )
         if (ijob .eq. -1) go to 900
*        *---end revcom matvec----------
          call swsub ( fwk(ib1),
     &    b, fwk(ib1),  iparm , fparm , iwk , fwk , ier  )
          call swdot (fwk(istab+ird), fwk(ib1),
     &    fwk(ib1),  iparm , fparm , iwk , fwk , ier  )
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.1).or.(iparm(13).eq.3))) go to 5642
                 go to 5643
 5642  continue
          call swset (fwk(ib2),fwk(ib1),
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5644
 5643 continue
          iva = ib1 
          ivql = ib2 
         ireq = 5
        iretlb = 2037
        go to 1500
 2037   continue
         if (ijob .eq. -1) go to 900
 5644 continue
*        *---end revcom precon----------
          call swdot (fwk(istab+ilrd ), fwk(ib2),
     &    fwk(ib2),  iparm , fparm , iwk , fwk , ier  )
          call swset (fwk(iu1),fwk(ib2),
     &     iparm , fparm , iwk , fwk , ier  )
*        *---begin revcom precon----------
         if (.not. ((iparm(13).eq.2).or.(iparm(13).eq.3))) go to 5645
                 go to 5646
 5645  continue
          call swset (fwk(iu2),fwk(iu1),
     &     iparm , fparm , iwk , fwk , ier  )
                go to 5647
 5646 continue
          ivql = iu1 
          ivqr = iu2 
         ireq = 7
        iretlb = 2038
        go to 1500
 2038   continue
         if (ijob .eq. -1) go to 900
 5647 continue
*        *---end revcom precon----------
          call swdot (fwk(istab+irlrd), fwk(iu2),
     &    fwk(iu2),  iparm , fparm , iwk , fwk , ier  )
        if (iparm(19) .eq. 1) then
          if (fwk(istab1+ied) .eq. 0e0) then
                ier = 4
                call xersho ( ier, 'scgnpr' , iparm ,
     &                        'Value of RELERR may be inaccurate' )
          else
            fparm(8) = sqrt(fwk(istab2+ied)/fwk(istab1+ied))
          endif
        endif
        if (fwk(istab1+ird) .eq. 0e0) then
                ier = 4
                call xersho ( ier, 'scgnpr' , iparm ,
     &                        'Value of RELRSD may be inaccurate' )
        else
          fparm(7) = sqrt(fwk(istab2+ird)/fwk(istab1+ird))
        endif
        if (iparm(2) .ge. 4) then
        write (iparm(1),'()') 
        write (iparm(1),'(1x, 60(''*''))') 
        write (iparm(1),'()') 
        write (iparm(1),'(5x,''Final iteration statistics'')') 
        write (iparm(1),'()') 
        if (iparm(19) .eq. 1) then
        write (iparm(1),'(5x,''2-norm of uexact            '',1pg20.10)'
     &) sqrt(fwk(istab+iud))
        endif
        write (iparm(1),'(5x,''2-norm of b                 '',1pg20.10)'
     &) sqrt(fwk(istab+ibd))
        write (iparm(1),'(5x,''2-norm of Ql*b              '',1pg20.10)'
     &) sqrt(fwk(istab+ilbd))
        write (iparm(1),'(5x,''2-norm of Qr*Ql*b           '',1pg20.10)'
     &) sqrt(fwk(istab+irlbd))
        if (iparm(19) .eq. 1) then
        write (iparm(1),'(5x,''2-norm of error             '',1pg20.10)'
     &) sqrt(fwk(istab+ied))
        endif
        write (iparm(1),'(5x,''2-norm of residual          '',1pg20.10)'
     &) sqrt(fwk(istab+ird))
        write (iparm(1),'(5x,''2-norm of Ql*r              '',1pg20.10)'
     &) sqrt(fwk(istab+ilrd))
        write (iparm(1),'(5x,''2-norm of Qr*Ql*r           '',1pg20.10)'
     &) sqrt(fwk(istab+irlrd))
        if (iparm(19) .eq. 1) then
        write (iparm(1),'(5x,''Relative 2-norm of error    '',1pg20.10)'
     &) fparm(8)
        endif
        write (iparm(1),'(5x,''Relative 2-norm of residual '',1pg20.10)'
     &) fparm(7)
        write (iparm(1),'()') 
        write (iparm(1),'(1x, 60(''*''))') 
        write (iparm(1),'()') 
        endif
        call sffre (iu2,iparm(3)*(1), iparm , fparm , iwk , fwk , ier )
        call sffre (iu1,iparm(3)*(1), iparm , fparm , iwk , fwk , ier )
        call sffre (ib2,iparm(3)*(1), iparm , fparm , iwk , fwk , ier )
        call sffre (ib1,iparm(3)*(1), iparm , fparm , iwk , fwk , ier )
        call sffre (istab2,8, iparm , fparm , iwk , fwk , ier )
        call sffre (istab1,8, iparm , fparm , iwk , fwk , ier )
                go to 5635
 5634 continue
 5635 continue
*     *---end iteration statistics calculation----------
*
      call sffre (idel,iparm(3)*(1), iparm , fparm , iwk , fwk , ier )
      call sffre (iv3,iparm(3)*(1), iparm , fparm , iwk , fwk , ier )
      call sffre (iv2,iparm(3)*(1), iparm , fparm , iwk , fwk , ier )
      call sffre (iv1,iparm(3)*(1), iparm , fparm , iwk , fwk , ier )
      if (.not.((iparm(13).eq.2).or.(iparm(13).eq.3))) then
      else
        if (.not.((iparm(13).eq.1).or.(iparm(13).eq.3))) then
        else
          call sffre (iqrqlr,iparm(3)*(1), iparm , fparm , iwk , fwk , i
     &er )
        endif
      endif
      call sffre (ir ,iparm(3)*(1), iparm , fparm , iwk , fwk , ier )
      call sffre (iap,iparm(3)*(1), iparm , fparm , iwk , fwk , ier )
      call sffre (ip ,iparm(3)*(1), iparm , fparm , iwk , fwk , ier )
*
*
*
*
      call sffre (iftab,12, iparm , fparm , iwk , fwk , ier )
      call xifre (iitab,30, iparm , fparm , iwk , fwk , ier )
*
      if (itimer .eq. 1) then
        call xtimer (tc2,tr2,iparm(22),0)
        fparm(1) = fparm(1) + (tc2-tc1)
        fparm(2) = fparm(2) + (tr2-tr1)
      endif
*
 910  continue
      if (iparm(27) .eq. 0) then
        call sifppr ( -1 , iparm, fparm, ier )
      endif
*
*     *---reset iclev----------
      if (iparm(27) .eq. 0) iparm(27) = 0
      iparm(11) = (iparm(6))
      ireq        = -1
      return
*
*-------------------------------brancher--------------------------------
*
*
 1000 continue
*
      ipme  = 0
      nproc = 1
      iitab = iparm(11)
      if (iitab.eq.(iparm(6)) .and. ijob.eq.-1) go to 910
*
      if (iparm(11) .eq. (iparm(6))) then
                ier = -4
                call xersho ( ier, 'scgnpr' , iparm ,
     &                        'Values of ijob and iparm(iptr) incompatib
     &le' )
      endif
*
      call xtimer (tc1,tr1,iparm(22),0)
*
*     *---remember parameters out of tables----------
        iitab  = iwk(iitab+0)
        iftab  = iwk(iitab+1)
        istab  = iwk(iitab+2)
        iwkstg = iwk(iitab+3)
        iclstg = iwk(iitab+4)
        in     = iwk(iitab+5)
        is     = iwk(iitab+6)
        itimer = iwk(iitab+7)
        iretlb = iwk(iitab+8)
        ivitmp = iwk(iitab+9)
        ivotmp = iwk(iitab+10)
        iv1tst = iwk(iitab+11)
        iv2tst = iwk(iitab+12)
        istab1 = iwk(iitab+13)
        istab2 = iwk(iitab+14)
        ib1    = iwk(iitab+15)
        ib2    = iwk(iitab+16)
        iu1    = iwk(iitab+17)
        iu2    = iwk(iitab+18)
        ir     = iwk(iitab+19)
        ip     = iwk(iitab+20)
        iap    = iwk(iitab+21)
        iqlr   = iwk(iitab+22)
        iqrqlr = iwk(iitab+23)
        iv1    = iwk(iitab+24)
        iv2    = iwk(iitab+25)
        iv3    = iwk(iitab+26)
        idel   = iwk(iitab+27)
        ns     = iwk(iitab+28)
        i      = iwk(iitab+29)
        top = fwk(iftab+0)
        bot = fwk(iftab+1)
        alphan = fwk(iftab+2)
        pap    = fwk(iftab+3)
        rnorm  = fwk(iftab+4)
        r0norm = fwk(iftab+5)
        rr     = fwk(iftab+6)
        rrold  = fwk(iftab+7)
        vlamda = fwk(iftab+8)
        alpha  = fwk(iftab+9)
        cval   = fwk(iftab+10)
        rdot   = fwk(iftab+11)
*
      if (ijob.ne.3 .and. ijob.ne.-1) then
                ier = -4
                call xersho ( ier, 'scgnpr' , iparm ,
     &                        'Argument ijob invalid' )
      endif
*
*     *---perform the branch to where we left off----------
      go to (2000,2001,2002,2003,2004,2005,2006,2007,2008,2009,2010,2011
     &,2012,2013,2014,2015,2016,2017,2018,2019,2020,2021,2022,2023,2024,
     &2025,2026,2027,2028,2029,2030,2031,2032,2033,2034,2035,2036,2037,2
     &038),
     &       iretlb - 2000 + 1
*
*---------------------------request of caller---------------------------
*
*
 1500 continue
      iparm(11) = iitab
*     *---save tables----------
        iwk(iitab+0) = iitab 
        iwk(iitab+1) = iftab 
        iwk(iitab+2) = istab 
        iwk(iitab+3) = iwkstg
        iwk(iitab+4) = iclstg
        iwk(iitab+5) = in    
        iwk(iitab+6) = is    
        iwk(iitab+7) = itimer
        iwk(iitab+8) = iretlb
        iwk(iitab+9) = ivitmp
        iwk(iitab+10) = ivotmp
        iwk(iitab+11) = iv1tst
        iwk(iitab+12) = iv2tst
        iwk(iitab+13) = istab1
        iwk(iitab+14) = istab2
        iwk(iitab+15) = ib1   
        iwk(iitab+16) = ib2   
        iwk(iitab+17) = iu1   
        iwk(iitab+18) = iu2   
        iwk(iitab+19) = ir    
        iwk(iitab+20) = ip    
        iwk(iitab+21) = iap   
        iwk(iitab+22) = iqlr  
        iwk(iitab+23) = iqrqlr
        iwk(iitab+24) = iv1   
        iwk(iitab+25) = iv2   
        iwk(iitab+26) = iv3   
        iwk(iitab+27) = idel  
        iwk(iitab+28) = ns    
        iwk(iitab+29) = i     
        fwk(iftab+0) = top
        fwk(iftab+1) = bot
        fwk(iftab+2) = alphan
        fwk(iftab+3) = pap   
        fwk(iftab+4) = rnorm 
        fwk(iftab+5) = r0norm
        fwk(iftab+6) = rr    
        fwk(iftab+7) = rrold 
        fwk(iftab+8) = vlamda
        fwk(iftab+9) = alpha 
        fwk(iftab+10) = cval  
        fwk(iftab+11) = rdot  
      call xtimer (tc2,tr2,iparm(22),0)
      fparm(1) = fparm(1) + (tc2-tc1)
      fparm(2) = fparm(2) + (tr2-tr1)
      return
*
*
*--------------------------------end------------------------------------
*
  500 format ( 1x, a, 1pg20.10 )
*
      end
